//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>
#include "memblock.h"
#define MAXBLOCKSIZE 1024
#define MINBLOCKSIZE 256

//#define BOOL bool
//#define FALSE false
//#define TRUE true
//#define min(a,b) __min(a,b)
//#define max(a,b) __max(a,b)



MemFrag::MemFrag()
{
	memset((char *)this,0,sizeof(*this));
}
MemFrag::~MemFrag()
{
	if(Data!=NULL) delete Data;
}

BOOL MemFrag::Alloc(int size)
{
Data=new char[size];
if(!Data) return FALSE;
// TODO: REMOVE THIS - DEBUG ONLY
memset(Data,0xcd,size);

Size=size;
return TRUE;
}

MemBlock::MemBlock()
{
	memset((char *)this,0,sizeof(*this));
	Data=new char[MINBLOCKSIZE];
	if(!Data) return;
	// TODO: REMOVE THIS - DEBUG ONLY
	memset(Data,0xcd,MINBLOCKSIZE);

	Size=MINBLOCKSIZE;
	Current=Last=this;
}

MemBlock::~MemBlock()
{
	MemFrag *tmp;
	while(Next) {
		tmp=Next;
		Next=Next->Next;
		delete tmp;
	}
}


BOOL MemBlock::Append(int bytes,char *_Data)
{
	char *orgdata=_Data;
	MemFrag *tmp;
	// CHECK IF ENOUGH ROOM IN LAST BLOCK
int slack=Last->Size-Last->Used;

while(bytes>slack)
{
// GROW CHAIN
if(orgdata) memmove(Last->Data+Last->Used,_Data,slack);
Last->Used+=slack;
_Data+=slack;
bytes-=slack;
TotalUsed+=slack;

tmp=new MemFrag;
if(!tmp) return FALSE;
if(!tmp->Alloc(min(MAXBLOCKSIZE,((bytes+MINBLOCKSIZE-1)/MINBLOCKSIZE)*MINBLOCKSIZE))) { delete tmp; return FALSE; }

// ADD TO MAIN CHAIN
tmp->Prev=Last;
Last->Next=tmp;
tmp->Offset=Last->Offset+Last->Used;
Last=tmp;
slack=tmp->Size;

}

	// JUST COPY
	if(orgdata) memmove(Last->Data+Last->Used,_Data,bytes);
	Last->Used+=bytes;
	TotalUsed+=bytes;
	Current=Last;
	return TRUE;
}


void MemBlock::Shrink(int bytes)		// REMOVE BYTES AT THE END
{
	MemFrag *tmp;
	if(bytes<=0) return;
while(bytes>=Last->Used && Last!=this)
{
// REMOVE LAST BLOCK
tmp=Last;
bytes-=Last->Used;
TotalUsed-=Last->Used;
Last=Last->Prev;
Last->Next=NULL;
delete tmp;
}

if(Last==this) {
Used-=bytes;
TotalUsed-=bytes;
if(Used<0) { TotalUsed=Used=0; }
Current=Last;
return;
}

Last->Used-=bytes;
TotalUsed-=bytes;
Current=Last;
return;
}

MemFrag *MemBlock::GetFrag(int off)
{
// FAST FIND BLOCK
// SEE IF OFFSET LIES WITHIN USED MEMORY
if(off<Used) return this;
if(off>=Last->Offset) return Last;
if(off>=Current->Offset)
{
// SEE IF IS WHITHIN CURRENT FRAG
	if(off<Current->Offset+Current->Used) return Current;

// START SEARCHING FROM CURRENT UP
	MemFrag *tmp=Current->Next;
	while(tmp) {
		if(off<tmp->Offset+tmp->Used) return tmp;
		tmp=tmp->Next;
	}
	// SHOULD NEVER GET TO THIS POINT UNLESS SOMETHING IS BROKEN
	return Last;

}
// START SEARCH FROM CURRENT DOWN
	MemFrag *tmp=Current->Prev;
	while(tmp) {
		if(off>=tmp->Offset) return tmp;
		tmp=tmp->Prev;
	}
	// SHOULD NEVER GET TO THIS POINT UNLESS SOMETHING IS BROKEN
	return this;
}

void MemBlock::ShiftOffset(MemFrag *block,int shift)
{
	while(block) {
		block->Offset+=shift;
		block=block->Next;
	}
}

BOOL MemBlock::Insert(int _offset,int bytes,char *Data)
{
	if(_offset<0 || _offset>TotalUsed) return FALSE;
	if(bytes<=0) return TRUE;
MemFrag *ptr1=GetFrag(_offset);
MemFrag *ptr2=ptr1->Next;
int slack1,slack2;
slack1=ptr1->Size-ptr1->Used;
if(ptr2) slack2=ptr2->Size-ptr2->Used;
else slack2=0;

if(slack1+slack2<bytes) {
// ADD NEW BLOCK BETWEEN EXISTING BLOCKS
	int needed=bytes-slack1;
	slack2=0;
	MemFrag *tmp;
	
	while(needed>0) {		// INSERT AS MANY BLOCKS AS NEEDED
	tmp=new MemFrag;
	if(!tmp) return FALSE;
	if(!tmp->Alloc(min(MAXBLOCKSIZE,((needed+MINBLOCKSIZE-1)/MINBLOCKSIZE)*MINBLOCKSIZE))) { delete tmp; return FALSE; }

	// ADD TO MAIN CHAIN
	tmp->Prev=ptr1;
	ptr1->Next=tmp;
	if(ptr2) { tmp->Next=ptr2; ptr2->Prev=tmp; }
	else { tmp->Next=NULL; Last=tmp; }
	tmp->Offset=ptr1->Offset+ptr1->Used;
	slack2+=tmp->Size;
	needed-=tmp->Size;
	ptr2=tmp;
	}
}

// NOW IT IS GUARANTEED TO HAVE ENOUGH SLACK ON BOTH BLOCKS


	// MAKE ROOM ON SECOND BLOCK
	int move2=max(0,bytes-slack1);
	int newbytes=0,nmove1,nmove2=0;

	nmove1=ptr1->Offset+ptr1->Used-_offset;		// USED BYTES TO MOVE WITHIN BLOCK 1

	if(move2)
	{
		newbytes=max(0,bytes-(ptr1->Offset+ptr1->Size-_offset));	// BYTES TO STORE INTO SECOND BLOCK
		nmove2=move2-newbytes;						// USED BYTES TO MOVE BROM BLOCK 1 TO 2
		nmove1-=nmove2;
		char *Dataptr=Data+bytes-newbytes;
		int cnewbytes=newbytes;
		int cmove2=move2;
		while(cnewbytes>ptr2->Size) {
			// FILL ALL INTERMEDIATE BLOCKS WITH NEW DATA
			if(Data) memmove(ptr2->Data,Dataptr,ptr2->Size);
			ptr2->Used=ptr2->Size;
			ShiftOffset(ptr2->Next,ptr2->Size);
			slack2-=ptr2->Size;
			Dataptr+=ptr2->Size;
			cnewbytes-=ptr2->Size;
			cmove2-=ptr2->Size;
			ptr2=ptr2->Next;

		}

		memmove(ptr2->Data+cmove2,ptr2->Data,ptr2->Used);
		ptr2->Used+=cmove2;
		ShiftOffset(ptr2->Next,cmove2);

		if(Data) memmove(ptr2->Data,Dataptr,cnewbytes);
		memmove(ptr2->Data+cnewbytes,ptr1->Data+_offset-ptr1->Offset+nmove1,nmove2);
	}

	// DO THE MOVE
	memmove(ptr1->Data+_offset-ptr1->Offset+bytes,ptr1->Data+_offset-ptr1->Offset,nmove1);
	if(Data) memmove(ptr1->Data+_offset-ptr1->Offset,Data,bytes-newbytes);
	ptr1->Used+=bytes-move2;
	ShiftOffset(ptr1->Next,bytes-move2);
	TotalUsed+=bytes;
	return TRUE;
}

void MemBlock::Delete(int _offset,int bytes)
{
	// REMOVE BYTES AT SPECIFIED LOCATION
	if(_offset<0) return;
	if(_offset>TotalUsed) return;
	if(bytes<=0) return;

	MemFrag *ptr=GetFrag(_offset);
	int tmp;
	int end=_offset+bytes;
	if(end>TotalUsed) { end=TotalUsed; bytes=end-_offset; }
	while(end>ptr->Offset+ptr->Used) {
		tmp=ptr->Used;
		ptr->Used=_offset-ptr->Offset;
		_offset=ptr->Offset+tmp;
		if(ptr->Used==0) {
			// REMOVE BLOCK
			if(ptr!=this) {
				if(ptr->Prev) { ptr->Prev->Next=ptr->Next; }
				if(ptr->Next) { ptr->Next->Prev=ptr->Prev; }
				if(Last==ptr) Last=ptr->Prev;
				MemFrag *fragtmp=ptr;
				ptr=ptr->Next;
				delete fragtmp;
			} else ptr=ptr->Next;
		} else ptr=ptr->Next;
	}

	// MOVE LAST BLOCK
	memmove(ptr->Data+_offset-ptr->Offset,ptr->Data+end-ptr->Offset,ptr->Offset+ptr->Used-end);
	ptr->Used-=end-_offset;
	if(ptr->Offset>end-bytes) ptr->Offset=end-bytes;
	ShiftOffset(ptr->Next,-bytes);
	TotalUsed-=bytes;
	Current=ptr;
	if(ptr->Used==0) {
			if(ptr!=this) {
				if(ptr->Prev) { ptr->Prev->Next=ptr->Next; Current=ptr->Prev; }
				if(ptr->Next) { ptr->Next->Prev=ptr->Prev; Current=ptr->Next; }
				if(Last==ptr) Last=ptr->Prev;
				delete ptr;
			}
	} else Consolidate(ptr);
	// ELIMINATE EMPTY FIRST BLOCKS
	if(Used==0 && Next) {
				delete Data;
				Data=Next->Data;
				Size=Next->Size;
				Used=Next->Used;
				Next->Used=0;
				Next->Data=NULL;
				ptr=Next;
				if(ptr->Prev) { ptr->Prev->Next=ptr->Next; }
				if(ptr->Next) { ptr->Next->Prev=ptr->Prev; }
				if(Last==ptr) Last=ptr->Prev;
				if(Current==ptr) Current=this;
				delete ptr;
			}

}


void MemBlock::Consolidate(MemFrag *ptr)
{
if(!ptr->Next) return;

if(ptr->Size>=ptr->Used+ptr->Next->Used) {
// CONSOLIDATE BOTH BLOCKS INTO FIRST
memmove(ptr->Data+ptr->Used,ptr->Next->Data,ptr->Next->Used);
ptr->Used+=ptr->Next->Used;
ptr=ptr->Next;
if(ptr->Prev) { ptr->Prev->Next=ptr->Next; }
if(ptr->Next) { ptr->Next->Prev=ptr->Prev; }
if(Last==ptr) Last=ptr->Prev;
if(Current==ptr) Current=ptr->Prev;
delete ptr;
return;
}
if(ptr->Next->Size>=ptr->Used+ptr->Next->Used) {
if(ptr==this) return;
// CONSOLIDATE BOTH BLOCKS INTO NEXT
memmove(ptr->Next->Data+ptr->Used,ptr->Next->Data,ptr->Next->Used);
memmove(ptr->Next->Data,ptr->Data,ptr->Used);
ptr->Next->Used+=ptr->Used;
ptr->Next->Offset=ptr->Offset;
ptr=ptr->Prev;
if(ptr->Prev) { ptr->Prev->Next=ptr->Next; }
if(ptr->Next) { ptr->Next->Prev=ptr->Prev; }
if(Last==ptr) Last=ptr->Next;
if(Current==ptr) Current=ptr->Next;
delete ptr;
return;
}

}


int MemBlock::WastedSpace()
{
int size=0;
MemFrag *ptr=this;

while(ptr)
{
	size+=ptr->Size;
	ptr=ptr->Next;
}

return size-TotalUsed;
}


int MemBlock::GetLowerLimit(int offset)
{
MemFrag *ptr=GetFrag(offset);
return ptr->Offset;
}

int MemBlock::GetUpperLimit(int offset)
{
MemFrag *ptr=GetFrag(offset);
return ptr->Offset+ptr->Used-1;
}


void MemBlock::MemMove(int destoffset,int srcoffset, int bytes)	// INTERNAL MEMORY MOVEMENT
{
// TRUNCATE AS NEEDED
	if(destoffset<0) return;
	if(srcoffset<0) return;
	if(destoffset+bytes>TotalUsed) bytes=max(TotalUsed-destoffset,0);
	if(srcoffset+bytes>TotalUsed) bytes=max(TotalUsed-srcoffset,0);
	if(bytes<=0) return;

		MemFrag *ptr; 
		int movebytes;

	if(srcoffset>destoffset) {

		// SELECT DIRECTION OF COPY: UPWARDS
		ptr=GetFrag(srcoffset);

		while(bytes) {
			movebytes=min(bytes,ptr->Used-(srcoffset-ptr->Offset));
			MemMoveUp(destoffset,ptr->Data+srcoffset-ptr->Offset,movebytes);
			destoffset+=movebytes;
			bytes-=movebytes;
			srcoffset+=movebytes;
			ptr=ptr->Next;
		}
		return;
	}

		// SELECT DIRECTION OF COPY: DOWNWARDS
		srcoffset+=bytes;
		destoffset+=bytes;
		ptr=GetFrag(srcoffset-1);
		while(bytes) {
			movebytes=min(bytes,srcoffset-ptr->Offset);
			MemMoveDn(destoffset,ptr->Data+srcoffset-ptr->Offset,movebytes);
			destoffset-=movebytes;
			bytes-=movebytes;
			srcoffset-=movebytes;
			ptr=ptr->Prev;
		}
		return;
}

void MemBlock::MemMoveUp(int destoffset,char *src,int bytes)	// INTERNAL USE
{
	MemFrag *ptr=GetFrag(destoffset);
	int movebytes;

	while(bytes) {
		movebytes=min(bytes,ptr->Used-(destoffset-ptr->Offset));
		memmove(ptr->Data+destoffset-ptr->Offset,src,movebytes);
		destoffset+=movebytes;
		bytes-=movebytes;
		src+=movebytes;
		ptr=ptr->Next;
	}
}

void MemBlock::MemMoveDn(int destoffset,char *src,int bytes)	// INTERNAL USE
{
	MemFrag *ptr=GetFrag(destoffset-1);
	int movebytes;
		while(bytes) {
			movebytes=min(bytes,destoffset-ptr->Offset);
			destoffset-=movebytes;
			src-=movebytes;
			memmove(ptr->Data+destoffset-ptr->Offset,src,movebytes);
			bytes-=movebytes;
			ptr=ptr->Prev;
		}

}

void MemBlock::MemMove(int destoffset,MemBlock &src,int srcoffset,int bytes)	// FROM EXTERNAL MemBlock OBJECT
{
if(&src==this) MemMove(destoffset,srcoffset,bytes);	// USE INTERNAL FUNCTION IF BOTH ARE THE SAME BLOCK

	if(destoffset<0) return;
	if(srcoffset<0) return;
	if(destoffset+bytes>TotalUsed) bytes=max(TotalUsed-destoffset,0);
	if(srcoffset+bytes>src.TotalUsed) bytes=max(src.TotalUsed-srcoffset,0);
	if(bytes<=0) return;

		MemFrag *ptr; 
		int movebytes;

		ptr=src.GetFrag(srcoffset);

		while(bytes) {
			movebytes=min(bytes,ptr->Used-(srcoffset-ptr->Offset));
			MemMoveUp(destoffset,ptr->Data+srcoffset-ptr->Offset,movebytes);
			destoffset+=movebytes;
			bytes-=movebytes;
			srcoffset+=movebytes;
			ptr=ptr->Next;
		}

}

void MemBlock::MemMove(int destoffset,char *src,int bytes) 	// FROM EXTERNAL DATA BLOCK, DO NOT USE FOR INTERNAL MOVEMENTS
{
	if(destoffset<0) return;
	if(destoffset+bytes>TotalUsed) bytes=max(TotalUsed-destoffset,0);
	if(bytes<=0) return;

	MemMoveUp(destoffset,src,bytes);

}

void MemBlock::MemMove(char *dest,int srcoffset,int bytes)	// COPY TO EXTERNAL DATA BLOCK, DO NOT USE FOR INTERNAL TRANSFERS
{
	if(srcoffset<0) return;
	if(srcoffset+bytes>TotalUsed) bytes=max(TotalUsed-srcoffset,0);
	if(bytes<=0) return;

	MemFrag *ptr=GetFrag(srcoffset);
	int movebytes;

	while(bytes) {
		movebytes=min(bytes,ptr->Used-(srcoffset-ptr->Offset));
		memmove(dest,ptr->Data+srcoffset-ptr->Offset,movebytes);
		srcoffset+=movebytes;
		bytes-=movebytes;
		dest+=movebytes;
		ptr=ptr->Next;
	}

}
