//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***




#include <hpgcc3.h>


// CONVERT AN XREAL TO A STRING GIVEN THE FLAGS
// XREAL IS EXPECTED TO BE IN BASE-10 IF data->base==10
// OR BASE-2 IN ANY OTHER BASE


char *xr_xrtoa(XREAL *number, struct conversion_data *data)
{

char tempbuf[65];
char *numbuff,*ptr,*pad0,*truncate;
int totaldigits,intdigits,fracdigits,postdigits=0,trail=0;

totaldigits=xr_getdigits(number,tempbuf,data->base);

if(data->base==10) intdigits=totaldigits+number->Exp;
else {
	int bits=xr_getnumbits(data->base);
	int bias=(64+number->Exp)%bits;
	if(bias<0) bias+=bits;
	if(bias) bias=bits-bias;
	intdigits=(64+number->Exp+bias)/bits;
}

if(data->type==12) {
if(!(data->flags&32)) trail=1;
// DECIDE WHETHER TO USE f OR e FORMAT
if(intdigits<-4 || intdigits>data->precision) data->type=4;
else data->type=8;
}
else {
//if(data->type==4) trail=1;
}

fracdigits=0;
if(data->type==4) intdigits=1;
if(data->type&12) fracdigits=data->precision;

if(fracdigits+intdigits>totaldigits) {
	if(data->type==8) postdigits=fracdigits+intdigits-totaldigits;
	if(intdigits>totaldigits) { intdigits=totaldigits; fracdigits=0; }
	else { fracdigits=totaldigits-intdigits; }
}

// ALLOCATE NUMBUFF
numbuff=(char *)malloc(((intdigits<1)? 1:intdigits)+fracdigits+postdigits+50);
if(!numbuff) return NULL;



// READY FOR OUTPUT
ptr=numbuff;
// 1=LEFT ALIGN, 2=FORCE SIGN, 4=0 PADD, 8=BLANK, 16=PREFIX, 32=FORCE DECIMAL
if(data->type!=2) {
	if(number->Sign) *ptr++='-'; else {
		if(data->flags&10) *ptr++=(data->flags&2)? '+':' ';
	}
}
if(data->flags&16) {
if(data->base==8 || data->base==16) *ptr++='0';
if(data->base==16) *ptr++='x';
}

pad0=ptr;

// OUTPUT INTEGER DIGITS

if(intdigits+fracdigits>0 && intdigits+fracdigits<totaldigits)
{
	// ROUND LAST PRINTED DIGIT
	if(xr_getbasedigit(tempbuf[intdigits+fracdigits],data->base)>=(data->base>>1))
	{
		int j=intdigits+fracdigits-1;
		do {
			tempbuf[j]++;
			if(tempbuf[j]==':') tempbuf[j]+=39;
			if(xr_getbasedigit(tempbuf[j],1024)==data->base)  { tempbuf[j]='0'; j--; }
			else break;
		} while(j>=0);

	}
}

if(intdigits>0) { memcpy(ptr,tempbuf,intdigits); ptr+=intdigits; }
else *ptr++='0';
// OUTPUT FRACTIONAL PART
if(fracdigits || (data->flags&32)) *ptr++='.';
while(intdigits<0 && fracdigits>0) { *ptr++='0'; --fracdigits; ++intdigits; }
truncate=ptr;
if(fracdigits) {
		memcpy(ptr,tempbuf+intdigits,fracdigits);
		ptr+=fracdigits;
	}

while(postdigits--) *ptr++='0';


// REMOVE TRAILING ZEROS
if(trail) {
while(ptr!=truncate) {
	if(*(ptr-1)!='0') break;
	--ptr;
}
if((*(ptr-1)=='.') && !(data->flags&32)) --ptr;	// REMOVE DECIMAL POINT
}
// OUTPUT EXPONENT

if(data->type==4) {
int exponent=number->Exp+totaldigits-1;
*ptr++='e';
*ptr++=(exponent<0)? '-':'+';
XREAL xrtmp;
char expbuff[23];
int expdigits,k;
xrtmp.Mant=abs(exponent);
xrtmp.Exp=0;
expdigits=xr_getdigits10(&xrtmp,expbuff);
for(k=0;(k<expdigits-3) && expbuff[k]=='0';++k);
memcpy(ptr,expbuff+k,expdigits-k);
ptr+=expdigits-k;
}

*ptr=0;

// FINAL PRESENTATION
if(data->flags&64) strupr(numbuff);

// PADDING
int length=(int)(ptr-numbuff),width;
width=((length>data->width)? length:data->width);
char *finalbuff=(char *)malloc( width+1);
// INSUFFICIENT MEMORY TO PRINT
if(!finalbuff) { free(numbuff); return NULL; }

memset(finalbuff,((data->flags&5)==4)? '0':' ',width);
finalbuff[width]=0;

if((data->flags&5)!=4) pad0=numbuff;

memcpy(finalbuff,numbuff,pad0-numbuff);
memcpy(finalbuff+(pad0-numbuff)+ ((data->flags&1)? 0:(width-length) ),pad0,strlen(pad0));

free(numbuff);
return finalbuff;
}
