//& *** (c) 2006-2011 The HPGCC3 Team ***
//& Claudio Lapilli
//& Ingo Blank
//&
//& This file is licensed under the terms and conditions of the
//& HPGCC3 license that is included with the source distribution.
//& *** (c) 2006-2011 The HPGCC3 Team ***

#include <hpgcc3.h>

// $Id: sat3_common_idioms.c 601 2008-09-01 22:33:13Z claudio $

/*

Some Saturn stack convenience functions, which ease the use of common idioms
like pushing/popping standard C data types int/double/string.

< ibl ; 2007-10-21 >

*/

int
sat3_pop_int(int def)
{
    
    SAT_OBJECT sob = sat3_pop();
    while(sat3_istag(sob)) sob=sat3_striptag(sob);
    
    if (sat3_isreal(sob)) return sat3_convertreal_int(sob);
    if (sat3_iszint(sob)) return sat3_convertzint_int(sob);
    if (sat3_ishxs(sob)) return sat3_converthxs_int(sob);
    if (sat3_isbint(sob)) return sat3_convertbint_int(sob);
	
    return def;
}

BOOL
sat3_push_llong_zint(LONGLONG n)
{
    SAT_OBJECT sob ;
    if (!(sob = sat3_createzint_llong(ALLOC_TEMPOB,n))) return FALSE;
    return sat3_push(sob);
}


BOOL
sat3_push_int_zint(int n)
{
    return sat3_push_llong_zint((LONGLONG) n);
}

BOOL
sat3_push_int_real(int n)
{
    SAT_OBJECT sob ;
    if (!(sob = sat3_createreal_int(ALLOC_TEMPOB,n))) return FALSE;
    return sat3_push(sob);
}
BOOL
sat3_push_int_bint(int n)
{
    SAT_OBJECT sob ;
    if (!(sob = sat3_createbint_int(ALLOC_TEMPOB,n))) return FALSE;
    return sat3_push(sob);
}


#include <decContext.h>
#include <decNumberLocal.h>
#include <decNumber.h>
#define decNumberSize(digits) (2*sizeof(int32_t)+sizeof(uint8_t) + D2U((digits)) * sizeof(decNumberUnit))

decNumber      *
decNumberAlloc(int digits)
{
    return (decNumber *) calloc(1, decNumberSize(digits));
}

BOOL
sat3_push_decn_zint(decNumber *dcn)
{
    SAT_OBJECT sob ;
    decNumber *idcn;
    decContext set;
    decContextDefault(&set, DEC_INIT_BASE);
    set.traps = 0;
    idcn = decNumberAlloc(dcn->digits);
    set.digits = dcn->digits;
    decNumberToIntegralValue(idcn,dcn,&set);
    if (!(sob = sat3_createzint_decn(ALLOC_TEMPOB,idcn))) return FALSE;
    free(idcn);
    return sat3_push(sob);
}

decNumber *
sat3_pop_decn(char *def)
{
	decNumber *decn;
	decContext set;
	
	SAT_OBJECT sob = sat3_pop();
	while(sat3_istag(sob)) sob=sat3_striptag(sob);
	
	if (sat3_iszint(sob)) {
		if (! sat3_convertzint_decn(sob,&decn)) goto defVal;
	} else if (sat3_isreal(sob)) {
		decn = decNumberAlloc(16);
		if (! sat3_convertreal_decn(sob,decn)) {
			free(decn);
			goto defVal;
		}
	} else {
		
	defVal:
		
		decContextDefault(&set, DEC_INIT_BASE);
		set.traps = 0;
		
		decn = decNumberAlloc(strlen(def));
		decNumberFromString(decn,def,&set);
	}
	
	return decn;
}

double
sat3_pop_dbl(double def)
{
  SAT_OBJECT sob = sat3_pop();
  while(sat3_istag(sob)) sob=sat3_striptag(sob);

    if (sat3_isreal(sob)) return sat3_convertreal_dbl(sob);
    if (sat3_iszint(sob)) return sat3_convertzint_dbl(sob);
    if (sat3_ishxs(sob)) return (double)sat3_converthxs_int(sob);
    if (sat3_isbint(sob)) return (double)sat3_convertbint_int(sob);
    
    return def;
}    

BOOL
sat3_push_dbl_real(double x)
{
    SAT_OBJECT sob ;
    if (!(sob = sat3_createreal_dbl(ALLOC_TEMPOB,x))) return FALSE;
    return sat3_push(sob);
}

    
char *
sat3_pop_string_alloc()
{
    char *s = (char *) NULL;
    int strl;
    
    SAT_OBJECT sob = sat3_pop();
    while(sat3_istag(sob)) sob=sat3_striptag(sob);

    if(sat3_iscstr(sob)) {
	strl = sat3_strsize(sob);
	if((s = (char *)calloc(1,strl+1)))
		sat3_convertcstr_str(sob,s,strl+1);

    } 
    
    return s;
}

char *
sat3_pop_string(char *s, int strl)
{
    
    SAT_OBJECT sob = sat3_pop();
    while(sat3_istag(sob)) sob=sat3_striptag(sob);

    *s = '\0';
    if(sat3_iscstr(sob)) 
		sat3_convertcstr_str(sob,s,strl);    
    return s;
}




BOOL
sat3_push_string(const char *s)
{
    
    SAT_OBJECT sob ;
    if (!(sob =sat3_createcstr_str(ALLOC_TEMPOB,(char *)s))) return FALSE;
    return sat3_push(sob);
	
}




